extends Object

const LINE_COMENT : String = "//"
const FORWARD_SLASH : String = "/"

const LINE_SEPARATOR : String = "\n"

const BASE : String = "base"
const DASH : String = "-"
const COMMAND : String = "command"
const LINK : String = "link"
const SHORTCUT : String = "shortcut"
const TAG : String = "tag"
const EXPRESSION : String = "expression"
const ASSIGNMENT : String = "assignment"
const OPTION : String = "option"
const OR : String = "or"
const DESTINATION : String = "destination"

var WHITESPACE : String = "\\s*"

var _states : Dictionary = {}
var _defaultState : LexerState

var _currentState : LexerState

var _indentStack : Array = []
var _shouldTrackIndent : bool = false


func _init():
	create_states()

func create_states():
	var patterns : Dictionary = {}
	patterns[WolGlobals.TokenType.Text] = ".*"

	patterns[WolGlobals.TokenType.Number] = "\\-?[0-9]+(\\.[0-9+])?"
	patterns[WolGlobals.TokenType.Str] = "\"([^\"\\\\]*(?:\\.[^\"\\\\]*)*)\""
	patterns[WolGlobals.TokenType.TagMarker] = "\\#"
	patterns[WolGlobals.TokenType.LeftParen] = "\\("
	patterns[WolGlobals.TokenType.RightParen] =  "\\)"
	patterns[WolGlobals.TokenType.EqualTo] = "(==|is(?!\\w)|eq(?!\\w))"
	patterns[WolGlobals.TokenType.EqualToOrAssign] = "(=|to(?!\\w))"
	patterns[WolGlobals.TokenType.NotEqualTo] = "(\\!=|neq(?!\\w))"
	patterns[WolGlobals.TokenType.GreaterThanOrEqualTo] = "(\\>=|gte(?!\\w))"
	patterns[WolGlobals.TokenType.GreaterThan] = "(\\>|gt(?!\\w))"
	patterns[WolGlobals.TokenType.LessThanOrEqualTo] = "(\\<=|lte(?!\\w))"
	patterns[WolGlobals.TokenType.LessThan] = "(\\<|lt(?!\\w))"
	patterns[WolGlobals.TokenType.AddAssign] =  "\\+="
	patterns[WolGlobals.TokenType.MinusAssign] = "\\-="
	patterns[WolGlobals.TokenType.MultiplyAssign] = "\\*="
	patterns[WolGlobals.TokenType.DivideAssign] = "\\/="
	patterns[WolGlobals.TokenType.Add] = "\\+"
	patterns[WolGlobals.TokenType.Minus] = "\\-"
	patterns[WolGlobals.TokenType.Multiply] = "\\*"
	patterns[WolGlobals.TokenType.Divide] = "\\/"
	patterns[WolGlobals.TokenType.Modulo] = "\\%"
	patterns[WolGlobals.TokenType.And] = "(\\&\\&|and(?!\\w))"
	patterns[WolGlobals.TokenType.Or] = "(\\|\\||or(?!\\w))"
	patterns[WolGlobals.TokenType.Xor] = "(\\^|xor(?!\\w))"
	patterns[WolGlobals.TokenType.Not] = "(\\!|not(?!\\w))"
	patterns[WolGlobals.TokenType.Variable] = "\\$([A-Za-z0-9_\\.])+"
	patterns[WolGlobals.TokenType.Comma] = "\\,"
	patterns[WolGlobals.TokenType.TrueToken] = "true(?!\\w)"
	patterns[WolGlobals.TokenType.FalseToken] = "false(?!\\w)"
	patterns[WolGlobals.TokenType.NullToken] = "null(?!\\w)"
	patterns[WolGlobals.TokenType.BeginCommand] = "\\<\\<"
	patterns[WolGlobals.TokenType.EndCommand] = "\\>\\>"
	patterns[WolGlobals.TokenType.OptionStart] = "\\[\\["
	patterns[WolGlobals.TokenType.OptionEnd] = "\\]\\]"
	patterns[WolGlobals.TokenType.OptionDelimit] = "\\|"
	patterns[WolGlobals.TokenType.Identifier] = "[a-zA-Z0-9_:\\.]+"
	patterns[WolGlobals.TokenType.IfToken] = "if(?!\\w)"
	patterns[WolGlobals.TokenType.ElseToken] = "else(?!\\w)"
	patterns[WolGlobals.TokenType.ElseIf] = "elseif(?!\\w)"
	patterns[WolGlobals.TokenType.EndIf] = "endif(?!\\w)"
	patterns[WolGlobals.TokenType.Set] = "set(?!\\w)"
	patterns[WolGlobals.TokenType.ShortcutOption] = "\\-\\>\\s*"

	#compound states
	var shortcut_option : String= SHORTCUT + DASH + OPTION
	var shortcut_option_tag : String = shortcut_option + DASH + TAG
	var command_or_expression : String= COMMAND + DASH + OR + DASH + EXPRESSION
	var link_destination : String = LINK + DASH + DESTINATION

	_states = {}

	_states[BASE] = LexerState.new(patterns)
	_states[BASE].add_transition(WolGlobals.TokenType.BeginCommand,COMMAND,true)
	_states[BASE].add_transition(WolGlobals.TokenType.OptionStart,LINK,true)
	_states[BASE].add_transition(WolGlobals.TokenType.ShortcutOption,shortcut_option)
	_states[BASE].add_transition(WolGlobals.TokenType.TagMarker,TAG,true)
	_states[BASE].add_text_rule(WolGlobals.TokenType.Text)

	_states[TAG] = LexerState.new(patterns)
	_states[TAG].add_transition(WolGlobals.TokenType.Identifier,BASE)

	_states[shortcut_option] = LexerState.new(patterns)
	_states[shortcut_option].track_indent = true
	_states[shortcut_option].add_transition(WolGlobals.TokenType.BeginCommand,EXPRESSION,true)
	_states[shortcut_option].add_transition(WolGlobals.TokenType.TagMarker,shortcut_option_tag,true)
	_states[shortcut_option].add_text_rule(WolGlobals.TokenType.Text,BASE)
	
	_states[shortcut_option_tag] = LexerState.new(patterns)
	_states[shortcut_option_tag].add_transition(WolGlobals.TokenType.Identifier,shortcut_option)

	_states[COMMAND] = LexerState.new(patterns)
	_states[COMMAND].add_transition(WolGlobals.TokenType.IfToken,EXPRESSION)
	_states[COMMAND].add_transition(WolGlobals.TokenType.ElseToken)
	_states[COMMAND].add_transition(WolGlobals.TokenType.ElseIf,EXPRESSION)
	_states[COMMAND].add_transition(WolGlobals.TokenType.EndIf)
	_states[COMMAND].add_transition(WolGlobals.TokenType.Set,ASSIGNMENT)
	_states[COMMAND].add_transition(WolGlobals.TokenType.EndCommand,BASE,true)
	_states[COMMAND].add_transition(WolGlobals.TokenType.Identifier,command_or_expression)
	_states[COMMAND].add_text_rule(WolGlobals.TokenType.Text)

	_states[command_or_expression] = LexerState.new(patterns)
	_states[command_or_expression].add_transition(WolGlobals.TokenType.LeftParen,EXPRESSION)
	_states[command_or_expression].add_transition(WolGlobals.TokenType.EndCommand,BASE,true)
	_states[command_or_expression].add_text_rule(WolGlobals.TokenType.Text)

	_states[ASSIGNMENT] = LexerState.new(patterns)
	_states[ASSIGNMENT].add_transition(WolGlobals.TokenType.Variable)
	_states[ASSIGNMENT].add_transition(WolGlobals.TokenType.EqualToOrAssign, EXPRESSION)
	_states[ASSIGNMENT].add_transition(WolGlobals.TokenType.AddAssign, EXPRESSION)
	_states[ASSIGNMENT].add_transition(WolGlobals.TokenType.MinusAssign, EXPRESSION)
	_states[ASSIGNMENT].add_transition(WolGlobals.TokenType.MultiplyAssign, EXPRESSION)
	_states[ASSIGNMENT].add_transition(WolGlobals.TokenType.DivideAssign, EXPRESSION)

	_states[EXPRESSION] = LexerState.new(patterns)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.EndCommand, BASE)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.Number)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.Str)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.LeftParen)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.RightParen)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.EqualTo)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.EqualToOrAssign)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.NotEqualTo)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.GreaterThanOrEqualTo)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.GreaterThan)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.LessThanOrEqualTo)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.LessThan)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.Add)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.Minus)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.Multiply)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.Divide)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.Modulo)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.And)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.Or)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.Xor)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.Not)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.Variable)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.Comma)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.TrueToken)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.FalseToken)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.NullToken)
	_states[EXPRESSION].add_transition(WolGlobals.TokenType.Identifier)

	_states[LINK] = LexerState.new(patterns)
	_states[LINK].add_transition(WolGlobals.TokenType.OptionEnd, BASE, true)
	_states[LINK].add_transition(WolGlobals.TokenType.OptionDelimit, link_destination, true)
	_states[LINK].add_text_rule(WolGlobals.TokenType.Text)

	_states[link_destination] = LexerState.new(patterns)
	_states[link_destination].add_transition(WolGlobals.TokenType.Identifier)
	_states[link_destination].add_transition(WolGlobals.TokenType.OptionEnd, BASE)

	_defaultState = _states[BASE]

	for stateKey in _states.keys():
		_states[stateKey].stateName = stateKey

	pass

func tokenize(text:String)->Array:
	
	_indentStack.clear()
	_indentStack.push_front(IntBoolPair.new(0,false))
	_shouldTrackIndent = false

	var tokens : Array  = []

	_currentState = _defaultState

	var lines : PoolStringArray = text.split(LINE_SEPARATOR)
	lines.append("")

	var lineNumber : int = 1

	for line in lines:
		tokens+=tokenize_line(line,lineNumber)
		lineNumber+=1

	var endOfInput : Token = Token.new(WolGlobals.TokenType.EndOfInput,_currentState,lineNumber,0)
	tokens.append(endOfInput)

	# print(tokens)

	return tokens

func tokenize_line(line:String, lineNumber : int)->Array:
	var tokenStack : Array = []

	var freshLine = line.replace("\t","    ").replace("\r","")

	#record indentation
	var indentation = line_indentation(line)
	var prevIndentation : IntBoolPair = _indentStack.front()

	if _shouldTrackIndent && indentation > prevIndentation.key:
		#we add an indenation token to record indent level
		_indentStack.push_front(IntBoolPair.new(indentation,true))

		var indent : Token = Token.new(WolGlobals.TokenType.Indent,_currentState,lineNumber,prevIndentation.key)
		indent.value = "%*s" % [indentation - prevIndentation.key,""]

		_shouldTrackIndent = false
		tokenStack.push_front(indent)

	elif indentation < prevIndentation.key:
		#de-indent and then emit indentaiton token

		while indentation < _indentStack.front().key:
			var top : IntBoolPair = _indentStack.pop_front()
			if top.value:
				var deIndent : Token = Token.new(WolGlobals.TokenType.Dedent,_currentState,lineNumber,0)
				tokenStack.push_front(deIndent)
	
	
	var column : int = indentation

	var whitespace : RegEx = RegEx.new()
	var error = whitespace.compile(WHITESPACE)
	if error != OK:
		printerr("unable to compile regex WHITESPACE")
		return []
	
	while column < freshLine.length():

		if freshLine.substr(column).begins_with(LINE_COMENT):
			break
		
		var matched : bool = false

		for rule in _currentState.rules:
			var found : RegExMatch = rule.regex.search(freshLine, column)
			
			if !found:
				continue

			var tokenText : String

			if rule.tokenType == WolGlobals.TokenType.Text:
				#if this is text then we back up to the most recent
				#delimiting token and treat everything from there as text.
				
				var startIndex : int = indentation

				if tokenStack.size() > 0 :
					while tokenStack.front().type == WolGlobals.TokenType.Identifier:
						tokenStack.pop_front()
					
					var startDelimitToken : Token = tokenStack.front()
					startIndex =  startDelimitToken.column

					if startDelimitToken.type == WolGlobals.TokenType.Indent:
						startIndex += startDelimitToken.value.length()
					if startDelimitToken.type == WolGlobals.TokenType.Dedent:
						startIndex = indentation
				#
				
				column = startIndex
				var endIndex : int = found.get_start() + found.get_string().length()

				tokenText = freshLine.substr(startIndex,endIndex-startIndex)
			
			else:
				tokenText = found.get_string()

			column += tokenText.length()

			#pre-proccess string
			if rule.tokenType == WolGlobals.TokenType.Str:
				tokenText = tokenText.substr(1,tokenText.length() - 2)
				tokenText = tokenText.replace("\\\\", "\\")
				tokenText = tokenText.replace("\\\"","\"")

			var token : Token = Token.new(rule.tokenType,_currentState,lineNumber,column,tokenText)
			token.delimitsText = rule.delimitsText

			tokenStack.push_front(token)

			if rule.enterState != null && rule.enterState.length() > 0:

				if !_states.has(rule.enterState):
					printerr("State[%s] not known - line(%s) col(%s)"%[rule.enterState,lineNumber,column])
					return []
				
				enter_state(_states[rule.enterState])

				if _shouldTrackIndent:
					if _indentStack.front().key < indentation:
						_indentStack.append(IntBoolPair.new(indentation,false))
			
			matched = true
			break

		if !matched:
			# TODO: Send out some helpful messages
			printerr("expectedTokens [%s] - line(%s) col(%s)"%["refineErrors.Lexer.tokenize_line",lineNumber,column])
			return []

		var lastWhiteSpace : RegExMatch = whitespace.search(line,column)
		if lastWhiteSpace:
			column += lastWhiteSpace.get_string().length()
		
	
	tokenStack.invert()

	return tokenStack

func line_indentation(line:String)->int:
	var indentRegex : RegEx = RegEx.new()
	indentRegex.compile("^(\\s*)")

	var found : RegExMatch = indentRegex.search(line)
	
	if !found || found.get_string().length() <= 0:
		return 0

	return found.get_string().length()

func enter_state(state:LexerState):
	_currentState = state;
	if _currentState.track_indent:
		_shouldTrackIndent = true

class Token:
	var type : int
	var value : String

	var lineNumber : int
	var column : int
	var text : String

	var delimitsText : bool= false
	var paramCount : int
	var lexerState : String

	func _init(type:int,state: LexerState, lineNumber:int = -1,column:int = -1,value:String =""):
		self.type = type
		self.lexerState = state.stateName
		self.lineNumber = lineNumber
		self.column = column
		self.value = value

	func _to_string():
		return "%s (%s) at %s:%s (state: %s)" % [WolGlobals.token_type_name(type),value,lineNumber,column,lexerState]
	

class LexerState:

	var stateName : String
	var patterns : Dictionary
	var rules : Array = []
	var track_indent : bool = false

	func _init(patterns):
		self.patterns = patterns

	func add_transition(type : int, state : String = "",delimitText : bool = false)->Rule:
		var pattern = "\\G%s" % patterns[type]
		# print("pattern = %s" % pattern)
		var rule = Rule.new(type,pattern,state,delimitText)
		rules.append(rule)
		return rule
	
	func add_text_rule(type : int, state : String = "")->Rule:
		if contains_text_rule() :
			printerr("State already contains Text rule")
			return null
		
		var delimiters:Array = []
		for rule in rules:
			if rule.delimitsText:
				delimiters.append("%s" % rule.regex.get_pattern().substr(2))

		var pattern = "\\G((?!%s).)*" % [PoolStringArray(delimiters).join("|")]
		var rule : Rule = add_transition(type,state)
		rule.regex = RegEx.new()
		rule.regex.compile(pattern)
		rule.isTextRule = true
		return rule

	func contains_text_rule()->bool:
		for rule in rules:
			if rule.isTextRule:
				return true
		return false
	

class Rule:
	var regex : RegEx

	var enterState : String
	var tokenType : int
	var isTextRule : bool
	var delimitsText : bool

	func _init(type : int , regex : String, enterState : String, delimitsText:bool):
		self.tokenType = type
		self.regex = RegEx.new()
		self.regex.compile(regex)
		self.enterState = enterState
		self.delimitsText = delimitsText

	func _to_string():
		return "[Rule : %s - %s]" % [WolGlobals.token_type_name(tokenType),regex]

class IntBoolPair:
	var key : int
	var value : bool

	func _init(key:int,value:bool):
		self.key = key
		self.value = value
	
